World:import("LB_lib.lua")
LB:installMainMenuHandler()
LB:clickThroughMinigames()
LB:installRandomDecisionHandler()

-- Mio Nishizono
-- Adapted from http://gamerssquare.kanpaku.jp/littlebusters.htm#r5
decisions = {
    -- May 14th
    "しばらく見守る",
    "よくない",
    "よくない",
    "よくない",
    "よくない",
    -- May 15th
    "貸す",
    "探しに行く",
    "しばらく見守る",
    -- May 16th
    "鈴を追う",
    "しばらく見守る",
    "ＱＲコードを携帯で読み込んでおく",
    "引き受ける",
    "巻き込まれる",
    "練習を続ける",
    "見つけるのを手伝う",
    -- May 17th
    "探してみる",
    "黙ってみてる",
    "携帯サイトにアクセスする",
    "次を読む", --　※ランダムで内容が変わる。好きなだけ”次を読む”を選んでも可
    "自習に戻る",
    "鈴を追う",
    "しばらく見守る",
    "屋上にいく",
    --※佐々美と対戦
    -- May 18th
    "鈴",
    "しばらく見守る", -- ※配膳ミッション
    "中庭に残る",
    "何訊いてたの？",
    "ジュースを買いに行く",
    "引き受ける",
    "やめさせる",
    -- ※謎の生命体戦。敗北でOK
    -- May 19th
    "鈴を探す",
    "しばらく見守る",
    "やっぱり気になる",
    "表に出る",
    "手伝う",
    "首をつっこむ",
    "中庭",
    "練習を続ける",
    "手伝う",
    -- May 20th
    "いく",
    "来ヶ谷",
    "ふわふわストロベリードーナツ",
    -- May 21st
    "しっかり読む",
    "真人と遊ぶ",
    -- 5/21
    "探してみる",
    "善意",
    "何を話しているか聞きにいく",
    "葉留佳",
    -- 5/22
    "掃除用具の整理", --※表示されない
    "呼びにいかない", --※表示されない
    "真人と遊ぶ",
    "中庭にいく",
    "恭介を追う",
    "何を話しているか聞きにいく",
    "クドに頷いた",
    "いく",
    -- May 23rd
    "クドに話しかける",
    "しばらく見守る",
    "やめさせる",
    "中庭にいく",
    "真人と遊ぶ",
    "何を話しているか聞きにいく",
    "学食に急ぐ",
    "やめておく",
    "恭介の部屋にいく",
    "みんなを呼ぶ",
    -- May 24th
    "真人と遊ぶ",
    "中庭にいく",
    "教室に戻る",
    "何を話しているか聞きにいく",
    "美魚",
    "クド",
    -- May 25th
    "これで決定",
    "廊下を走り抜ける",
    "断る",
    "終わりにする",
    "ジョー",
    "エイリアン",
    -- May 27th
    "また晴れの日にすればいいんじゃないかな",
    -- May 28th
    "中庭にいく",
    -- "鈴",
    -- "小毬",
    -- "葉留佳",
    -- "クド",
    -- "来ヶ谷",
    "美鳥",
    "善意"

    -- TODO: Need to inject a click here.?
}

World:setDecisionList(decisions)
