World:import("Bride_lib.lua")
Bride:installMainMenuHandler()

decisions = {
    "１日目から始める",
    -- 1st Tale
    "兀突骨大王？",
    "親しみを込めてカッちゃん？",
    "じゃ、葛城クン？",
    -- 3rd Tale
    "①　物のない様子。全てを失った様子。",
    "葛城と一緒に炊飯器のスイッチを入れる係",
    "櫻見さん、逢坂さんとポークのオレンジソースを作る",
    -- 6th Tale
    "宿題は自分の力でやるべき",
    "今度だけ見せてあげたら？",
    -- 7th Tale
    "受けない",
    -- 8th Tale
    "受ける",
    "牛丼屋",
    "愛生ちゃんと遊ぶ",
    "とにかく聖ちゃんを脱出させる",
    "③　子供の遊戯。けんけん。",
    "愛生ちゃんとお菓子を買いに",
    "だしの量が微妙に…",
    "聖ちゃんと愛生ちゃんの部屋に行く",
    "キュウリを食べられるようになろう",
    "あわてて電話を切る",
    "先を譲る",
    -- 9th Tale
    "遊園地",
    "聖ちゃんと愛生ちゃんの部屋に行く",
    "愛生ちゃんはお母さんのこと好き？",
    "黙っている",
    -- 10th Tale
    "聖ちゃんと愛生ちゃんの部屋に行く",
    -- 11th Tale
    "受ける",
    -- 12th Tale
    "今回だけだよ",
    -- 13th Tale
    "愛生ちゃん",
    "渡す"
}

World:setDecisionList(decisions)
