World:import("Air_lib.lua")
Air:installMainMenuHandler("DREAM")

-- Misuzu's route
decisions = {
    -- July 18
    "人形を探し歩く",
    "歩き出す",
    "話してみる",
    "正直に答える",
    "大人しくしておく",
    "友達に電話をかけてみる",
    "素直に頷く",
    "正直に答える",
    "飲める",
    "冗談じゃない",
    -- July 19
    "学校の中に入る",
    "なんでもない振りをする",
    "戸棚",
    "まともに答える",
    "手伝う",
    "「かわいいな」",
    "見せない",
    "聞く",
    "お仕置きってなんだ",
    "眠いので寝る",
    -- July 20
    "パン",
    "地図の上へ",
    "とりあえず、声をかけてみる",
    "何でもいいから恩を売っておく",
    "あえて触れずにおく",
    "おごらない",
    "実践してみる",
    "武田商店",
    "人形芸を見せてやる",
    "たぶんそうなるだろう",
    "お断りだ",
    "断る",
    -- July 21
    "買い取ったら観鈴にやる",
    "ばっちりだ、と答える",
    "書き足してもらう",
    "観鈴を信じる",
    "その通りだ",
    -- July 22
    "わかったよ",
    "神尾家へ",
    "ああ、そうする",
    "寝てしまう",
    -- July 23
    "敢えてステゴザウルスＴシャツを借りる",
    -- ・・・「ＡＩＲ」編でがおがおＴシャツが見ることができます（笑）。
    "敢えて着続けてみる",
    "俺のおごりだ",
    "『お茶のお供に最適です』",
    "商店街",
    "中に入る",
    "大人しく待つ",
    "観鈴の部屋にいく",
    "試してみる",
    "遊ぶ",
    -- July 26
    "ここにいる",
    -- July 27
    "…わかった",
    -- July 29
    "抱きしめる",
    "体を求める",
}

World:setDecisionList(decisions)
