World:import("Air_lib.lua")
Air:installMainMenuHandler("DREAM")

-- Minagi's route
decisions = {
    -- July 18
    "別の手を考える",
    "走り出す",
    "黙っておく",
    "走り出す",
    "さらに少女に移す",
    "大人しくしておく",
    "大人しく待っておく",
    "反抗してみせる",
    "ごまかす",
    "飲めない",
    "晴子に絡む",
    -- July 19
    "ギブアップ",
    "道を戻る",
    "せがむ",
    "冷蔵庫",
    "無視する",
    "見せない",
    "聞く",
    "お仕置きってなんだ",
    "眠いので寝る",
    -- July 20
    "なわとび",
    "地図の右へ",
    "とりあえず、声をかけてみる",
    "何でもいいから恩を売っておく",
    "バンダナのことを訊いてみる",
    "教えてほしいぞ",
    "無視",
    "やめておく",
    "橋",
    "人形芸を見せてやる",
    "たぶんそうなるだろう",
    "考えておく",
    "断る",
    -- July 21
    "考えておく",
    "力自慢なところを見せつける",
    "書き足してもらう",
    "自力で家を回ることにする",
    "断じて違う",
    -- July 22
    "胸を貸してやる",
    "豪快に投げ飛ばす",
    "ヒマを持て余している",
    "人形芸を見せてやる",
    "学校へ",
    "嫌いじゃない",
    "散歩に出かける",
    "先に進んでみる",
    "様子を見る",
    "どうにか説明する",
    -- July 23
    "濡れているのでいい",
    "商店街",
    "診療所の中",
    "意味もなく間を埋める",
    "駅のほうへ歩いていく",
    "拾って、返してやる",
    "やっちゃう",
    -- July 24
    "それだけはダメだ",
    "やめておく",
    -- July 25
    "手を貸す",
    "人形を渡す",
    "暑いからやめようと懇願する",
    "上がってこいと言う",
    "バンダナにはもう触れない",
    -- July 26
    "金を稼ぎに商店街へ",
    -- July 27
    "素直に受け取る",
    -- July 29
    "タ行",
    "それでも辞書",
    "嫉妬してみる",
    -- Aug 7
    "明日まで我慢してくれ"
}

World:setDecisionList(decisions)
