World:import("LB_lib.lua")
LB:installMainMenuHandler()
LB:clickThroughMinigames()

-- Yuiko's path. Copied and reworked from
-- http://gamerssquare.kanpaku.jp/littlebusters.htm#r2
decisions = {
    -- May 14th
    "しばらく見守る",
    "よくない",
    "よくない",
    "よくない",
    "よくない",
    -- May 15th
    "貸す",
    "探しに行く",
    "しばらく見守る",
    -- May 16th
    "鈴を追う",
    "しばらく見守る",
    "ＱＲコードを携帯で読み込んでおく",
    "引き受ける",
    "巻き込まれる",
    "練習を続ける",
    "見つけるのを手伝う",
    -- May 17th
    "探してみる",
    "黙ってみてる",
    "携帯サイトにアクセスする",
    "次を読む", --　※ランダムで内容が変わる。好きなだけ”次を読む”を選んでも可
    "自習に戻る",
    "鈴を追う",
    "しばらく見守る",
    "屋上にいく",
    --※佐々美と対戦
    -- May 18th
    "鈴",
    "しばらく見守る", -- ※配膳ミッション
    "中庭に残る",
    "何訊いてたの？",
    "ジュースを買いに行く",
    "引き受ける",
    "やめさせる",
    -- ※謎の生命体戦。敗北でOK
    -- May 19th
    "鈴を探す",
    "しばらく見守る",
    "やっぱり気になる",
    "表に出る",
    "手伝う",
    "首をつっこむ",
    "中庭",
    "練習を続ける",
    "手伝う",
    -- May 20th
    "いく",
    "来ヶ谷",
    "チョコストライプのドーナツ",
    -- May 21st
    "しっかり読む",
    "ジュースを買いに行く",
    "練習に向かう",
    "練習を続ける",
    "来ヶ谷",
    -- May 22nd
    "ここが僕の家",
    "呼びにいく",
    "真人と遊ぶ",
    "裏庭にいく",
    "耳を澄ませてみる",
    "恭介を追う",
    "何を話しているか聞きにいく",
    "クドに首を振った",
    "いく",
    -- May 23rd
    "教室に入る",
    "しばらく見守る",
    "やめさせる",
    "学食にいく",
    "断る",
    "ジュースを買いに行く",
    "何を話しているか聞きにいく",
    "放っておく",
    "学食に急ぐ",
    "おかわりする",
    "恭介の部屋にいく",
    "みんなを呼ぶ",
    -- May 24th
    "ジュースを買いに行く",
    "見送る",
    "教室に戻る",
    "何を話しているか聞きにいく",
    "来ヶ谷",
    "鈴",
    -- May 25th
    "これで決定",
    "潜む",
    "断る",
    "練習を続ける",
    "練習を続ける",
    "ジョー",
    "エイリアン",
    -- May 28th
    "話さない",
    -- 来ヶ谷
    "しらを切る",
    "小毬",
    "クド",
    "葉留佳",
    "恭介",
    "あえて前",
    "それでも…",
    "その名前を覚えていない"
}

World:setDecisionList(decisions)
