;; Project specific local variables for emacs23. `compile-command' is the only
;; one not set by the individual files. Hopefully this list will grow.
((nil . ((compile-command . "scons")
         (tab-width . 2)
         (indent-tabs-mode . nil))))
